#ifndef MODULE_ASCII_TO_EBCDIC
#define MODULE_ASCII_TO_EBCDIC
/*****************************************************************************/
/*  MODULE NAME:  ascii2ebcdic.c                        MODULE TYPE:  (app)  */
/*****************************************************************************/
/*  MODULE IMPORTS:                                                          */
/*****************************************************************************/
#include <stdio.h>                                  /*  MODULE TYPE:  (sys)  */
#include <string.h>                                 /*  MODULE TYPE:  (sys)  */
/*****************************************************************************/
/*  MODULE INTERFACE:                                                        */
/*****************************************************************************/
/*

    This application is a typical UNIX filter designed to convert
    from ASCII to EBCDIC and vice-versa in a REVERSIBLE way WITHOUT
    LOSS of information. The filter supports different EBCDIC code
    pages, e.g. for OS/390 and BS2000.

    Input comes from standard input, output goes to standard output.

    Both can be redirected to/from files using the corresponding UNIX
    redirection operators.

    Available command line options are:

        -v     reverse the conversion (use inverse translation table)
        -o     use the EBCDIC code page for OS/390 (DEFAULT)
        -b     use the EBCDIC code page for BS2000
        -n     "neutral" or "no operation" - simply pass through all
               characters (overrides "-v", "-o" and "-b")

        -vo    use OS/390 EBCDIC code page, reverse conversion
        -ov    use OS/390 EBCDIC code page, reverse conversion
        -vb    use BS2000 EBCDIC code page, reverse conversion
        -bv    use BS2000 EBCDIC code page, reverse conversion

        -rev   reverse the conversion
        -os3   use OS/390 EBCDIC code page (DEFAULT)
        -bs2   use BS2000 EBCDIC code page
        -nop   simply pass through all characters

        -dump  dump the internal translation table (can be combined
               with "-v" or "-rev" to dump the inverse translation
               table) instead of performing any conversion (override)
        -init  initialize the translation table first before dumping
               it (this option has no effect if "-dump" is not present)

    The last options on the command line take precedence over the first ones,
    where applicable.

    "-v" and "-rev" are toggle switches, i.e., specifying this option twice
    is the same as not specifying it at all.

*/
/*****************************************************************************/
/*  MODULE RESOURCES:                                                        */
/*****************************************************************************/

/*
    Only characters that are visually the same or very similar in both
    the source and the target character set need to be specified here,
    the rest of this translation table is filled automatically.
*/

#define CHARSETS 2

/*
    Character set #0:
            EBCDIC code table for OS/390 (DEFAULT)
    Character set #1:
            EBCDIC code table for BS2000
*/

static int ascii2ebcdic[CHARSETS][0x100] =
{
    {
        0x00, 0x01, 0x02, 0x03,   0x37, 0x2D, 0x2E, 0x2F,  /* 00 */
        0x16, 0x05, 0x25, 0x0B,   0x0C, 0x0D, 0x0E, 0x0F,

        0x10, 0x11, 0x12, 0x13,   0x3C, 0x3D, 0x32, 0x26,  /* 10 */
        0x18, 0x19, 0x3F, 0x27,   0x1C, 0x1D, 0x1E, 0x1F,

        0x40, 0x4F, 0x7F, 0x7B,   0x5B, 0x6C, 0x50, 0x7D,  /* 20 */
        0x4D, 0x5D, 0x5C, 0x4E,   0x6B, 0x60, 0x4B, 0x61,

        0xF0, 0xF1, 0xF2, 0xF3,   0xF4, 0xF5, 0xF6, 0xF7,  /* 30 */
        0xF8, 0xF9, 0x7A, 0x5E,   0x4C, 0x7E, 0x6E, 0x6F,

        0xB5, 0xC1, 0xC2, 0xC3,   0xC4, 0xC5, 0xC6, 0xC7,  /* 40 */
        0xC8, 0xC9, 0xD1, 0xD2,   0xD3, 0xD4, 0xD5, 0xD6,

        0xD7, 0xD8, 0xD9, 0xE2,   0xE3, 0xE4, 0xE5, 0xE6,  /* 50 */
        0xE7, 0xE8, 0xE9, 0x63,   0xEC, 0xFC, 0x5F, 0x6D,

        0x79, 0x81, 0x82, 0x83,   0x84, 0x85, 0x86, 0x87,  /* 60 */
        0x88, 0x89, 0x91, 0x92,   0x93, 0x94, 0x95, 0x96,

        0x97, 0x98, 0x99, 0xA2,   0xA3, 0xA4, 0xA5, 0xA6,  /* 70 */
        0xA7, 0xA8, 0xA9, 0x43,   0xBB, 0xDC, 0x59, 0x07,

        0x9F, 0x04, 0x51, 0x42,   0x20, 0x44, 0x47, 0x48,  /* 80 */
        0x52, 0x53, 0x54, 0x57,   0x56, 0x58, 0x3B, 0x67,

        0x71, 0x9C, 0x9E, 0xCB,   0x75, 0xCD, 0xDB, 0xDD,  /* 90 */
        0xDF, 0x9B, 0xCF, 0x70,   0xB1, 0x80, 0xBF, 0xFF,

        0x45, 0x55, 0xCE, 0xDE,   0x49, 0x69, 0x9A, 0x7C,  /* A0 */
        0xAB, 0xAF, 0xBA, 0xB8,   0xB7, 0xAA, 0x8A, 0x8B,

        0x2B, 0x2C, 0x09, 0x21,   0x28, 0x65, 0x62, 0x64,  /* B0 */
        0xB4, 0x38, 0x31, 0x34,   0x33, 0xB0, 0xB2, 0x24,

        0x22, 0x17, 0x29, 0x06,   0x4A, 0x2A, 0x46, 0x66,  /* C0 */
        0x1A, 0x35, 0x08, 0x39,   0x36, 0x30, 0x3A, 0x68,

        0x8C, 0xAC, 0x72, 0x73,   0x74, 0x0A, 0xE0, 0x76,  /* D0 */
        0x77, 0x23, 0x15, 0x14,   0x5A, 0xCC, 0x78, 0xA1,

        0xEE, 0xE1, 0xEB, 0xED,   0xC0, 0xEF, 0xA0, 0x8E,  /* E0 */
        0xAE, 0xFE, 0xFB, 0xFD,   0x8D, 0xAD, 0xBC, 0xBE,

        0xCA, 0x8F, 0x1B, 0xB9,   0xB6, 0xFA, 0x6A, 0x9D,  /* F0 */
        0x90, 0xBD, 0xB3, 0xDA,   0xD0, 0xEA, 0x3E, 0x41
    },
    {
        0x00, 0x01, 0x02, 0x03,   0x37, 0x2D, 0x2E, 0x2F,  /* 00 */
        0x16, 0x05, 0x08, 0x0B,   0x0C, 0x0D, 0x0E, 0x0F,

        0x10, 0x11, 0x12, 0x13,   0x14, 0x3D, 0x32, 0x27,  /* 10 */
        0x18, 0x19, 0x3F, 0x1B,   0x1C, 0x1D, 0x1E, 0x1F,

        0x40, 0x5A, 0x7F, 0x7B,   0x5B, 0x6C, 0x50, 0x7D,  /* 20 */
        0x4D, 0x5D, 0x5C, 0x4E,   0x6B, 0x60, 0x4B, 0x61,

        0xF0, 0xF1, 0xF2, 0xF3,   0xF4, 0xF5, 0xF6, 0xF7,  /* 30 */
        0xF8, 0xF9, 0x7A, 0x5E,   0x4C, 0x7E, 0x6E, 0x6F,

        0x7C, 0xC1, 0xC2, 0xC3,   0xC4, 0xC5, 0xC6, 0xC7,  /* 40 */
        0xC8, 0xC9, 0xD1, 0xD2,   0xD3, 0xD4, 0xD5, 0xD6,

        0xD7, 0xD8, 0xD9, 0xE2,   0xE3, 0xE4, 0xE5, 0xE6,  /* 50 */
        0xE7, 0xE8, 0xE9, 0xBB,   0xBC, 0xBD, 0x6A, 0x6D,

        0x4A, 0x81, 0x82, 0x83,   0x84, 0x85, 0x86, 0x87,  /* 60 */
        0x88, 0x89, 0x91, 0x92,   0x93, 0x94, 0x95, 0x96,

        0x97, 0x98, 0x99, 0xA2,   0xA3, 0xA4, 0xA5, 0xA6,  /* 70 */
        0xA7, 0xA8, 0xA9, 0xFB,   0x4F, 0xFD, 0xFF, 0x07,

        0x04, 0x06, 0x15, 0x09,   0x0A, 0x1A, 0x21, 0x17,  /* 80 */
        0x22, 0x24, 0x20, 0x25,   0x23, 0x26, 0x28, 0x29,

        0x2A, 0x2B, 0x2C, 0x36,   0x30, 0x31, 0x33, 0x34,  /* 90 */
        0x35, 0x38, 0x39, 0x3C,   0x3A, 0x3B, 0x3E, 0x5F,

        0x41, 0xAA, 0xB0, 0xB1,   0x9F, 0xB2, 0xD0, 0xB5,  /* A0 */
        0x79, 0xB4, 0x9A, 0x8A,   0xBA, 0xCA, 0xAF, 0xA1,

        0x90, 0x8F, 0xEA, 0xFA,   0xBE, 0xA0, 0xB6, 0xB3,  /* B0 */
        0x9D, 0xDA, 0x9B, 0x8B,   0xB7, 0xB8, 0xB9, 0xAB,

        0x64, 0x65, 0x62, 0x66,   0x63, 0x67, 0x9E, 0x68,  /* C0 */
        0x74, 0x71, 0x72, 0x73,   0x78, 0x75, 0x76, 0x77,

        0xAC, 0x69, 0xED, 0xEE,   0xEB, 0xEF, 0xEC, 0xBF,  /* D0 */
        0x80, 0xE0, 0xFE, 0xDD,   0xFC, 0xAD, 0xAE, 0x59,

        0x44, 0x45, 0x42, 0x46,   0x43, 0x47, 0x9C, 0x48,  /* E0 */
        0x54, 0x51, 0x52, 0x53,   0x58, 0x55, 0x56, 0x57,

        0x8C, 0x49, 0xCD, 0xCE,   0xCB, 0xCF, 0xCC, 0xE1,  /* F0 */
        0x70, 0xC0, 0xDE, 0xDB,   0xDC, 0x8D, 0x8E, 0xDF
    }
};

/*****************************************************************************/
/*  MODULE IMPLEMENTATION:                                                   */
/*****************************************************************************/

static char *self;

static int ebcdic2ascii[CHARSETS][0x100];

void clear(int charset)
{
    int i;

    for ( i = 0x00; i < 0x100; i++ )
    {
        ebcdic2ascii[charset][i] = 0x00;
    }
}

void initialize(int charset)
{
    int i,j,k;

    clear(charset);
    for ( i = 0x00; i < 0x100; i++ )
    {
        if (ascii2ebcdic[charset][i] != 0x00)
        {
            ebcdic2ascii[charset][ascii2ebcdic[charset][i]] = i;
        }
    }
    for ( i = 0x00; i < 0x100; i++ )
    {
        if ((ascii2ebcdic[charset][i] == 0x00) &&
            (ebcdic2ascii[charset][i] == 0x00))
        {
            ascii2ebcdic[charset][i] = i;
            ebcdic2ascii[charset][i] = i;
        }
    }
    k = 0x00;
    for ( i = 0x00; i < 0x100; i++ )
    {
        if (ascii2ebcdic[charset][i] == 0x00)
        {
            for ( j = k; j < 0x100; j++ )
            {
                if (ebcdic2ascii[charset][j] == 0x00)
                {
                    ascii2ebcdic[charset][i] = j;
                    ebcdic2ascii[charset][j] = i;
                    k = j + 1;
                    break;
                }
            }
        }
    }
    for ( i = 0x00; i < 0x100; i++ )
    {
        if (((ascii2ebcdic[charset][i] == 0x00) ^ (i == 0x00)) ||
            ((ebcdic2ascii[charset][i] == 0x00) ^ (i == 0x00)))
        {
            fprintf(stderr,"%s: internal configuration error!\n", self);
            exit(1);
        }
    }
}

void dump_table(int init, int reverse)
{
    int charset,i,k;

    fprintf(stdout,"static int ascii2ebcdic[CHARSETS][0x100] =\n");
    fprintf(stdout,"{\n");
    for ( charset = 0; charset < CHARSETS; charset++ )
    {
        if (init) initialize(charset); else clear(charset);
        fprintf(stdout,"    {\n");
        for ( i = 0x00; i < 0x100; i++ )
        {
            k = i & 0x07;
            if      (k == 0x00) fprintf(stdout,"        ");
            else if (k == 0x04) fprintf(stdout,"  ");
            if (reverse) fprintf(stdout,"0x%02.2X", ebcdic2ascii[charset][i]);
            else         fprintf(stdout,"0x%02.2X", ascii2ebcdic[charset][i]);
            if (k == 0x07)
            {
                if (i == 0xFF)               fprintf(stdout,"\n");
                else if ((i & 0x0F) == 0x0F) fprintf(stdout,",\n\n");
                else                         fprintf(stdout,",  /* %02.2X */\n", i & 0xF0);
            }
            else                             fprintf(stdout,", ");
        }
        if (charset < CHARSETS-1) fprintf(stdout,"    },\n");
        else                      fprintf(stdout,"    }\n");
    }
    fprintf(stdout,"};\n");
}

int main(int argc, char *argv[])
{
    char *option;
    int   reverse = 0;
    int   charset = 0;
    int   neutral = 0;
    int   dump = 0;
    int   init = 0;
    int   err = 0;
    int   c;

    if ((self = strrchr(argv[0],'/')) != NULL) self++; else self = argv[0];
    for ( c = 1; c < argc; c++ )
    {
        option = argv[c];
        if (*option == '-')
        {
            option++;
            switch (strlen(option))
            {
                case 1:
                    if      (*option == 'v') reverse = ! reverse;
                    else if (*option == 'o') charset = 0;
                    else if (*option == 'b') charset = 1;
                    else if (*option == 'n') neutral = 1;
                    else err = 1;
                  break;
                case 2:
                    if      (strcmp(option,"vo") == 0)
                            { reverse = ! reverse; charset = 0; }
                    else if (strcmp(option,"ov") == 0)
                            { reverse = ! reverse; charset = 0; }
                    else if (strcmp(option,"vb") == 0)
                            { reverse = ! reverse; charset = 1; }
                    else if (strcmp(option,"bv") == 0)
                            { reverse = ! reverse; charset = 1; }
                    else err = 1;
                  break;
                case 3:
                    if      (strcmp(option,"rev") == 0) reverse = ! reverse;
                    else if (strcmp(option,"os3") == 0) charset = 0;
                    else if (strcmp(option,"bs2") == 0) charset = 1;
                    else if (strcmp(option,"nop") == 0) neutral = 1;
                    else err = 1;
                  break;
                case 4:
                    if      (strcmp(option,"dump") == 0) dump = 1;
                    else if (strcmp(option,"init") == 0) init = 1;
                    else err = 1;
                  break;
                default:
                    err = 1;
                  break;
            }
        }
        else err = 1;
        if (err)
        {
            fprintf(stderr,"%s: unknown option '%s'!\n", self, argv[c]);
            exit(1);
        }
    }
    if (dump) dump_table(init,reverse);
    else if (neutral)
    {
        while ((c = getchar()) != EOF) putchar(c);
    }
    else
    {
        initialize(charset);
        while ((c = getchar()) != EOF)
        {
            if (reverse) putchar(ebcdic2ascii[charset][c]);
            else         putchar(ascii2ebcdic[charset][c]);
        }
    }
    return(0);
}

/*****************************************************************************/
/*  VERSION:  1.0                                                            */
/*****************************************************************************/
/*  VERSION HISTORY:                                                         */
/*****************************************************************************/
/*                                                                           */
/*    01.09.2000  Version 1.0                                                */
/*                                                                           */
/*****************************************************************************/
/*  AUTHOR:                                                                  */
/*****************************************************************************/
/*                                                                           */
/*    Steffen Beyer                                                          */
/*    Ainmillerstr. 5 / App. 513                                             */
/*    D-80801 Munich                                                         */
/*    Germany                                                                */
/*                                                                           */
/*    mailto:sb@engelschall.com                                              */
/*    http://www.engelschall.com/u/sb/download/                              */
/*                                                                           */
/*****************************************************************************/
/*  COPYRIGHT:                                                               */
/*****************************************************************************/
/*                                                                           */
/*    Copyright (c) 2000 by Steffen Beyer.                                   */
/*    All rights reserved.                                                   */
/*                                                                           */
/*****************************************************************************/
/*  LICENSE:                                                                 */
/*****************************************************************************/
/*                                                                           */
/*    This program is free software; you can redistribute it and/or          */
/*    modify it under the terms of the GNU General Public License            */
/*    as published by the Free Software Foundation; either version 2         */
/*    of the License, or (at your option) any later version.                 */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the           */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the                          */
/*    Free Software Foundation, Inc.,                                        */
/*    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                  */
/*                                                                           */
/*    or download a copy from ftp://ftp.gnu.org/pub/gnu/COPYING-2.0          */
/*                                                                           */
/*****************************************************************************/
#endif
